//+------------------------------------------------------------------+
//|                                                          GNG.mqh |
//|                                             Copyright 2010, alsu |
//|                                                 alsufx@gmail.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2010, alsu"
#property link      "alsufx@gmail.com"

#include "Neurons.mqh"
//+------------------------------------------------------------------+
//| main class representing the algorithm itself                     |
//+------------------------------------------------------------------+
class CGNGAlgorithm
  {
public:
   //--- linked lists of the neuron and connection objects
   CGNGNeuronList   *Neurons;
   CGNGConnectionList *Connections;
   //--- algorithm parameters
   int               input_dimension;
   int               iteration_number;
   int               lambda;
   int               age_max;
   double            alpha;
   double            beta;
   double            eps_w;
   double            eps_n;
   int               max_nodes;

                     CGNGAlgorithm();
                    ~CGNGAlgorithm();
   virtual void      Init(int __input_dimension,
                          double &v1[],
                          double &v2[],
                          int __lambda,
                          int __age_max,
                          double __alpha,
                          double __beta,
                          double __eps_w,
                          double __eps_n,
                          int __max_nodes);
   virtual bool      ProcessVector(double &in[],bool train=true);
   virtual bool      StoppingCriterion();
  };
//+------------------------------------------------------------------+
//| constructor                                                      |
//+------------------------------------------------------------------+
CGNGAlgorithm::CGNGAlgorithm(void)
  {
   Neurons=new CGNGNeuronList();
   Connections=new CGNGConnectionList();
  }
//+------------------------------------------------------------------+
//| destructor                                                       |
//+------------------------------------------------------------------+
CGNGAlgorithm::~CGNGAlgorithm(void)
  {
   delete Neurons;
   delete Connections;
  }
//+------------------------------------------------------------------+
//| initializes algorithm with two samples and needed parameters     |
//| INPUT: v1,v2 - data samples                                      |
//|        __lambda - number of samples after which new nodes are    |
//|        inserted                                                  |
//|        __age_max - maximum age of the edges between neurons      |
//|        __alpha, __beta - used for errors adaptation              |
//|        __eps_w, __eps_n - used for weights adaptation            |
//|        __max_nodes - size limit for the neuronet                 |
//| OUTPUT: none                                                     |
//+------------------------------------------------------------------+
void CGNGAlgorithm::Init(int __input_dimension,
                         double &v1[],
                         double &v2[],
                         int __lambda,
                         int __age_max,
                         double __alpha,
                         double __beta,
                         double __eps_w,
                         double __eps_n,
                         int __max_nodes)
  {
   iteration_number=0;
   input_dimension=__input_dimension;
   lambda=__lambda;
   age_max=__age_max;
   alpha= __alpha;
   beta = __beta;
   eps_w = __eps_w;
   eps_n = __eps_n;
   max_nodes=__max_nodes;
   Neurons.Init(v1,v2);

   CGNGNeuron *tmp;
   tmp=Neurons.GetFirstNode();
   int uid1=tmp.uid;
   tmp=Neurons.GetLastNode();
   int uid2=tmp.uid;

   Connections.Init(uid1,uid2);
  }
//+------------------------------------------------------------------+
//| main function of the algorithm                                   |
//| INPUT: in - data sample                                          |
//|        train - run algorithm if true, otherwise just pass vector |
//|        to all the neurons (don't train the net)                  |
//| OUTPUT: true if stopping criterion matched, false otherwise      |
//+------------------------------------------------------------------+
bool CGNGAlgorithm::ProcessVector(double &in[],bool train=true)
  {
   if(ArraySize(in)!=input_dimension) return(StoppingCriterion());

   int i;

   CGNGNeuron *tmp=Neurons.GetFirstNode();
   while(CheckPointer(tmp))
     {
      tmp.ProcessVector(in);
      tmp=Neurons.GetNextNode();
     }

   if(!train) return(false);

   iteration_number++;

//--- Locate the two nodes nearest to in[], that is, the two nodes with
//--- reference vectors Ws and Wt such that ||Ws-in||^2 is the smallest
//--- value and ||Wt-in||^2 the second smallest, for all nodes.        
//--- ||*|| denotes Euclidean norm of the vector.                      
   CGNGNeuron *Winner,*SecondWinner;
   Neurons.FindWinners(Winner,SecondWinner);

//--- The winner-node must update its local error variable            
   Winner.E+=Winner.error;

//--- Move the winner and its topological neighbours (i.e. all nodes  
//--- connected to the winner by an edge) towards input vector by      
//--- fractions eps_w and eps_n of the distance.                       
   double delta[],weights[];

   Winner.Weights(weights);
   ArrayResize(delta,input_dimension);

   for(i=0;i<input_dimension;i++) delta[i]=eps_w*(in[i]-weights[i]);
   Winner.AdaptWeights(delta);

//--- Increment the age of all edges from node s to its topological    
//--- neighbours.                                                      
   CGNGConnection *tmpc=Connections.FindFirstConnection(Winner.uid);
   while(CheckPointer(tmpc))
     {
      if(tmpc.uid1==Winner.uid) tmp = Neurons.Find(tmpc.uid2);
      if(tmpc.uid2==Winner.uid) tmp = Neurons.Find(tmpc.uid1);

      tmp.Weights(weights);
      for(i=0;i<input_dimension;i++) delta[i]=eps_n*(in[i]-weights[i]);
      tmp.AdaptWeights(delta);

      tmpc.age++;

      tmpc=Connections.FindNextConnection(Winner.uid);
     }

//--- If winner and second winner are connected by an edge, then set   
//--- the age of that edge to 0. If they are not connected then create 
//--- an edge between them.                                            
   tmpc=Connections.Find(Winner.uid,SecondWinner.uid);
   if(tmpc) tmpc.age=0;
   else
     {
      Connections.Append();
      tmpc=Connections.GetLastNode();
      tmpc.uid1 = Winner.uid;
      tmpc.uid2 = SecondWinner.uid;
      tmpc.age=0;
     }

//--- If there are any edges with an age larger than age_max then      
//--- remove them. If, after this, there are nodes with no edges then  
//--- remove these nodes.                                              
   tmpc=Connections.GetFirstNode();
   while(CheckPointer(tmpc))
     {
      if(tmpc.age>age_max)
        {
         Connections.DeleteCurrent();
         tmpc=Connections.GetCurrentNode();
        }
      else tmpc=Connections.GetNextNode();
     }

   tmp=Neurons.GetFirstNode();
   while(CheckPointer(tmp))
     {
      if(!Connections.FindFirstConnection(tmp.uid))
        {
         Neurons.DeleteCurrent();
         tmp=Neurons.GetCurrentNode();
        }
      else tmp=Neurons.GetNextNode();
     }

//--- If the current iteration is an integer multiple of lambda and the
//--- maximum node-count has not been reached, then insert a new node r
   CGNGNeuron *u,*v;
   if(iteration_number%lambda==0 && Neurons.Total()<max_nodes)
     {

      //---   1.Find the node u with largest error.                          
      tmp=Neurons.GetFirstNode();
      u=tmp;
      while(CheckPointer(tmp=Neurons.GetNextNode()))
        {
         if(tmp.E>u.E)
            u=tmp;
        }

      //---   2.Among the neighbours of u, find the node v with the largest  
      //---     error.                                                       
      tmpc=Connections.FindFirstConnection(u.uid);
      if(tmpc.uid1==u.uid) v=Neurons.Find(tmpc.uid2);
      else v=Neurons.Find(tmpc.uid1);
      while(CheckPointer(tmpc=Connections.FindNextConnection(u.uid)))
        {
         if(tmpc.uid1==u.uid) tmp=Neurons.Find(tmpc.uid2);
         else tmp=Neurons.Find(tmpc.uid1);
         if(tmp.E>v.E)
            v=tmp;
        }

      //---   3.Insert the new node r in the "middle" between u and v.       
      double wr[],wu[],wv[];

      u.Weights(wu);
      v.Weights(wv);
      ArrayResize(wr,input_dimension);
      for(i=0;i<input_dimension;i++) wr[i]=(wu[i]+wv[i])/2;

      CGNGNeuron *r=Neurons.Append();
      r.Init(wr);

      //---   4.Create edges between u and r, and v and r, and then remove   
      //---     the edge between u and v.                                    
      tmpc=Connections.Append();
      tmpc.uid1=u.uid;
      tmpc.uid2=r.uid;

      tmpc=Connections.Append();
      tmpc.uid1=v.uid;
      tmpc.uid2=r.uid;

      Connections.Find(u.uid,v.uid);
      Connections.DeleteCurrent();

      //---   5.Decrease the error-variables of u and v and set the error of 
      //---     node r.                                                      
      u.E*=alpha;
      v.E*=alpha;
      r.E = u.E;
     }

//--- Decrease all error-variables of all nodes by a factor beta       
   tmp=Neurons.GetFirstNode();
   while(CheckPointer(tmp))
     {
      tmp.E*=(1-beta);
      tmp=Neurons.GetNextNode();
     }

//--- Check stoping criterion                                          
   return(StoppingCriterion());
  }
//+------------------------------------------------------------------+
//| stopping criterion of the algorithm. is dummy in this version of |
//| the file, i.e. returns false always                              |
//| INPUT: none                                                      |
//| OUTPUT: true if criterion matched, false otherwise               |
//+------------------------------------------------------------------+
bool CGNGAlgorithm::StoppingCriterion()
  {
   return(false);
  }
//+------------------------------------------------------------------+
//| Growing neural gas with utility factor                           |
//+------------------------------------------------------------------+
class CGNGUAlgorithm:public CGNGAlgorithm
  {
public:
   double            k;
   virtual void      Init(int __input_dimension,
                          double &v1[],
                          double &v2[],
                          int __lambda,
                          int __age_max,
                          double __alpha,
                          double __beta,
                          double __eps_w,
                          double __eps_n,
                          int __max_nodes,
                          double __k);
   virtual bool      ProcessVector(double &in[],bool train=true);
   virtual bool      StoppingCriterion();
  };
//+------------------------------------------------------------------+
//| initializes algorithm with two samples and needed parameters     |
//| INPUT: v1,v2 - data samples                                      |
//|        __lambda - number of samples after which new nodes are    |
//|        inserted                                                  |
//|        __age_max - maximum age of the edges between neurons      |
//|        __alpha, __beta - used for errors adaptation              |
//|        __eps_w, __eps_n - used for weights adaptation            |
//|        __max_nodes - size limit for the neuronet                 |
//|        __k - memory feature for non-stationary data              |
//| OUTPUT: none                                                     |
//+------------------------------------------------------------------+
void CGNGUAlgorithm::Init(int __input_dimension,
                          double &v1[],
                          double &v2[],
                          int __lambda,
                          int __age_max,
                          double __alpha,
                          double __beta,
                          double __eps_w,
                          double __eps_n,
                          int __max_nodes,
                          double __k)
  {
   iteration_number=0;
   input_dimension=__input_dimension;
   lambda=__lambda;
   age_max=__age_max;
   alpha= __alpha;
   beta = __beta;
   eps_w = __eps_w;
   eps_n = __eps_n;
   max_nodes=__max_nodes;
   k=__k;
   Neurons.Init(v1,v2);

   CGNGNeuron *tmp;
   tmp=Neurons.GetFirstNode();
   int uid1=tmp.uid;
   tmp=Neurons.GetLastNode();
   int uid2=tmp.uid;

   Connections.Init(uid1,uid2);
  }
//+------------------------------------------------------------------+
//| main function of the algorithm                                   |
//| INPUT: in - data sample                                          |
//|        train - run algorithm if true, otherwise just pass vector |
//|        to all the neurons (don't train the net)                  |
//| OUTPUT: true if stopping criterion matched, false otherwise      |
//+------------------------------------------------------------------+
bool CGNGUAlgorithm::ProcessVector(double &in[],bool train=true)
  {
   if(ArraySize(in)!=input_dimension) return(StoppingCriterion());

   int i;

   CGNGNeuron *tmp=Neurons.GetFirstNode();
   while(CheckPointer(tmp))
     {
      tmp.ProcessVector(in);
      tmp=Neurons.GetNextNode();
     }

   if(!train) return(false);

   iteration_number++;

//--- Locate the two nodes nearest to in[], that is, the two nodes with
//--- reference vectors Ws and Wt such that ||Ws-in||^2 is the smallest
//--- value and ||Wt-in||^2 the second smallest, for all nodes.        
//--- ||*|| denotes Euclidean norm of the vector.                      
   CGNGNeuron *Winner,*SecondWinner;
   Neurons.FindWinners(Winner,SecondWinner);

//--- The winner-node must update its local error variable and        
//--- utility factor                                                   
   Winner.E+=Winner.error;
   Winner.U+=SecondWinner.error-Winner.error;

//--- Move the winner and its topological neighbours (i.e. all nodes  
//--- connected to the winner by an edge) towards input vector by      
//--- fractions eps_w and eps_n of the distance.                       
   double delta[],weights[];

   Winner.Weights(weights);
   ArrayResize(delta,input_dimension);

   for(i=0;i<input_dimension;i++) delta[i]=eps_w*(in[i]-weights[i]);
   Winner.AdaptWeights(delta);

//--- Increment the age of all edges from node s to its topological    
//--- neighbours.                                                     
   CGNGConnection *tmpc=Connections.FindFirstConnection(Winner.uid);
   while(CheckPointer(tmpc))
     {
      if(tmpc.uid1==Winner.uid) tmp = Neurons.Find(tmpc.uid2);
      if(tmpc.uid2==Winner.uid) tmp = Neurons.Find(tmpc.uid1);

      tmp.Weights(weights);
      for(i=0;i<input_dimension;i++) delta[i]=eps_n*(in[i]-weights[i]);
      tmp.AdaptWeights(delta);

      tmpc.age++;

      tmpc=Connections.FindNextConnection(Winner.uid);
     }

//--- If winner and second winner are connected by an edge, then set   
//--- the age of that edge to 0. If they are not connected then create 
//--- an edge between them.                                            
   tmpc=Connections.Find(Winner.uid,SecondWinner.uid);
   if(tmpc) tmpc.age=0;
   else
     {
      Connections.Append();
      tmpc=Connections.GetLastNode();
      tmpc.uid1 = Winner.uid;
      tmpc.uid2 = SecondWinner.uid;
      tmpc.age=0;
     }

//--- If there are any edges with an age larger than age_max then      
//--- remove them. If, after this, the node with smallest utility has  
//--- lower than k-fraction of maximum error, remove it.               
   tmpc=Connections.GetFirstNode();
   while(CheckPointer(tmpc))
     {
      if(tmpc.age>age_max)
        {
         Connections.DeleteCurrent();
         tmpc=Connections.GetCurrentNode();
        }
      else tmpc=Connections.GetNextNode();
     }

   tmp=Neurons.GetFirstNode();
   double max_error;
   double min_U;
   CGNGNeuron *useless;

   if(CheckPointer(tmp))
     {
      max_error=tmp.error;
      min_U=tmp.U;
      useless=tmp;
     }
   while(CheckPointer(tmp=Neurons.GetNextNode()))
     {
      if(tmp.error>max_error)
        {
         max_error=tmp.error;
        }
      if(tmp.U<min_U)
        {
         min_U=tmp.U;
         useless=tmp;
        }
     }
   if(min_U!=0 && max_error/min_U>k) Neurons.Delete(Neurons.IndexOf(tmp));

//--- If the current iteration is an integer multiple of lambda and the
//--- maximum node-count has not been reached, then insert a new node r
   CGNGNeuron *u,*v;
   if(iteration_number%lambda==0 && Neurons.Total()<max_nodes)
     {

      //--- 1.Find the node u with largest error.                          
      tmp=Neurons.GetFirstNode();
      u=tmp;
      while(CheckPointer(tmp=Neurons.GetNextNode()))
        {
         if(tmp.E>u.E)
            u=tmp;
        }

      //--- 2.Among the neighbours of u, find the node v with the largest  
      //---   error.                                                       
      tmpc=Connections.FindFirstConnection(u.uid);
      if(tmpc.uid1==u.uid) v=Neurons.Find(tmpc.uid2);
      else v=Neurons.Find(tmpc.uid1);
      while(CheckPointer(tmpc=Connections.FindNextConnection(u.uid)))
        {
         if(tmpc.uid1==u.uid) tmp=Neurons.Find(tmpc.uid2);
         else tmp=Neurons.Find(tmpc.uid1);
         if(tmp.E>v.E)
            v=tmp;
        }

      //--- 3.Insert the new node r in the "middle" between u and v.       
      //--- Set its utility factor                                         
      double wr[],wu[],wv[];

      u.Weights(wu);
      v.Weights(wv);
      ArrayResize(wr,input_dimension);
      for(i=0;i<input_dimension;i++) wr[i]=(wu[i]+wv[i])/2;

      CGNGNeuron *r=Neurons.Append();
      r.Init(wr);

      r.U=(u.U+v.U)/2;
      //--- 4.Create edges between u and r, and v and r, and then remove   
      //---   the edge between u and v.                                    
      tmpc=Connections.Append();
      tmpc.uid1=u.uid;
      tmpc.uid2=r.uid;

      tmpc=Connections.Append();
      tmpc.uid1=v.uid;
      tmpc.uid2=r.uid;

      Connections.Find(u.uid,v.uid);
      Connections.DeleteCurrent();
      //--- 5.Decrease the error-variables of u and v and set the error of 
      //---   node r.                                                      
      u.E*=alpha;
      v.E*=alpha;
      r.E = u.E;
     }
//--- Decrease errors and utilities of all nodes by a factor beta      
   tmp=Neurons.GetFirstNode();
   while(CheckPointer(tmp))
     {
      tmp.E*=(1-beta);
      tmp.U*=(1-beta);
      tmp=Neurons.GetNextNode();
     }
//--- Check stoping criterion                                          
   return(StoppingCriterion());
  }
//+------------------------------------------------------------------+
//| stopping criterion of the algorithm. is dummy in this version of |
//| the file, i.e. returns false always                              |
//| INPUT: none                                                      |
//| OUTPUT: true if criterion matched, false otherwise               |
//+------------------------------------------------------------------+
bool CGNGUAlgorithm::StoppingCriterion()
  {
   return(false);
  }
//+------------------------------------------------------------------+
